#!/usr/bin/env python
# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

try:
    from gevent import monkey
    monkey.patch_all()
    SERVER = 'gevent'
except ImportError:
    SERVER = 'wsgiref'

import argparse
import base64
import binascii
from datetime import datetime
import hashlib
import json
import logging
import os

from bottle import request, response, route, run, default_app
from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader

from malcarve import scan
from malcarve.version import __version__

# Load templates
root = os.path.join(os.path.dirname(__file__), "html")
env = Environment()
env.loader = FileSystemLoader(root)
env.filters['b64encode'] = base64.b64encode
env.filters['b64decode'] = base64.b64decode
env.filters['utf8decode'] = lambda x: x.decode('utf8')
env.filters['hexlify'] = binascii.hexlify

jsondate = lambda obj: obj.isoformat() if isinstance(obj, datetime) else None
scanner = scan.Scanner()


def scan_file(infile, filename):
    """
    Scan the given data buffer and return a dictionary of results.
    """
    m = hashlib.md5()
    results = {'filename': filename,
               'status': 'Failed',
               'apiversion': __version__,
               'start': datetime.utcnow(),
               'carved': [],
               }
    try:
        # read whole thing into mem
        buf = infile.read()
        m.update(buf)
        results['md5'] = m.hexdigest()
        results['filesize'] = len(buf)
        # make human readable
        for r in scanner.scan_buffer(buf):
            if r.get('content'):
                r['content'] = base64.b64encode(r['content']).decode('utf-8')
                r['content_filename'] = scan.get_filename(filename, r)
            if r.get('content_with_overlay'):
                r['content_with_overlay'] = base64.b64encode(r['content_with_overlay']).decode('utf-8')
                r['content_with_overlay_filename'] = scan.get_filename(filename, r, True)
            if r.get('key'):
                r['key'] =  '0x' + binascii.hexlify(r['key']).decode('utf-8')
            if r.get('pattern'):
                r['pattern'] = repr(r['pattern'])[2:-1]
            results['carved'].append(r)
        # successfully processed but not necessarily carved anything
        results['status'] = 'Success'
    except Exception as ex:
        results['error'] = str(ex)
    finally:
        results['finish'] = datetime.utcnow()
        results['duration'] = (results['finish'] - results['start']).total_seconds()
        infile.close()
    return results

@route("/", name="home")
def home():
    """
    Main page, displays a submit file form.
    """
    template = env.get_template("submit.html")
    return template.render(base)


@route("/submit", method="POST", name="submit")
def submit_and_render():
    """
    Blocking POST handler for file submission.
    Runs malcarve on supplied file and returns results as rendered html.
    """
    data = request.files.file
    template = env.get_template("results.html")
    if not data:
        pass
    results = scan_file(data.file, data.filename)
    results.update(base)
    return template.render(results)

@route("/api/submit", method="POST", name="api_submit")
def api_submit():
    """
    Blocking POST handler for file submission.
    Runs malcarve on supplied file and returns results as json text.
    """
    data = request.files.file
    response.content_type = b'application/json'
    if not data or not hasattr(data, 'file'):
        return json.dumps({"status": "Failed", "stderr": "Missing form params"})
    return json.dumps(scan_file(data.file, data.filename), default=jsondate, indent=4)

@route("/api", name="api")
def api():
    """
    Display an api usage/help page.
    """
    template = env.get_template("api.html")
    return template.render(base)


def main():
    """
    Main entrypoint for command-line webserver.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-H", "--host", help="Web server Host address to bind to",
                        default="0.0.0.0", action="store", required=False)
    parser.add_argument("-p", "--port", help="Web server Port to bind to",
                        default=8080, action="store", required=False)
    args = parser.parse_args()

    logging.basicConfig()
    run(host=args.host, port=args.port, reloader=True, server=SERVER, debug=True)

# WSGI and template url support
application = default_app()
base = {'get_url': application.get_url}

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import binascii
from hashlib import md5
import importlib
from itertools import chain
import os
import quopri
import timeit

import begin
import progressbar

from malcarve.config import Config


def class_for_name(class_name):
    """
    Dynamically load the given class name.

    :param class_name: str with class name to import/load.
    :return: Loaded Class object.
    """
    module_name, class_name = class_name.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)

def get_filename(infile, carving, overlay=False):
    """
    Generate a suitable output filename for the carved object.
    
    :param infile: Input filename object was carved from
    :param carving: Dictionary of extracted object
    :param overlay: Is filename for saving with or without overlay included
    :return: String containing output filename
    """
    offset = '0x%08x' % carving['offset']
    fname = "%s.%s" % (os.path.basename(infile), offset)
    if carving.get('encoding'):
        fname += '.%s' % carving['encoding']
    if carving.get('stream_offset'):
        fname += '.0x%08x' % carving['stream_offset']
    fname += '.%s' % carving['scheme']
    if carving.get('key'):
        fname += '.%s' % binascii.hexlify(carving['key']).decode()
    if overlay:
        fname += '.with_overlay'
    return fname

class Scanner(object):
    """
    Main scanning object.
    
    Will derive configuration from default conf file locations unless
    overridden.
    """
    def __init__(self, config=None):
        self.config = config or Config()
        self.embedded_only = self.config.embedded_only

        self.deobfuscators = {}
        for x in self.config.deobfuscators.values():
            d = class_for_name(x.class_name)(x.name, x.description, x.schemes, x.modifiers, x.max_keysize, **x.other)
            self.deobfuscators[x.name] = d

        self.payloads = {}
        for x in self.config.payloads.values():
            p = class_for_name(x.class_name)(x.name, x.scan_min, x.scan_max, x.min_size, x.validation, **x.other)
            for d in x.deobs:
                p.add_scheme(self.deobfuscators[d])
            for pat in x.patterns:
                p.add_pattern(pat)
            self.payloads[p.name] = p

        self.decoders = {}
        for x in self.config.decoders.values():
            d = class_for_name(x.class_name)(x.name)
            self.decoders[x.name] = d

    def get_streams(self, stream, depth=0):
        """
        Recursively scan stream data, decoding additional streams to scan.

        :param stream: Input stream object to scan for additional streams.
        :param depth: How far has this recursed.
        :return: Generator of decoded streams to scan for obfuscated objects.
        """
        if not depth:
            stream = {
                'encoding': None,
                'offset': 0,
                'stream_id': 'file',
                'stream': stream,
                'parent': None,
            }
            yield stream

        if depth >= self.config.max_stream_depth:
            return

        for d in self.decoders.values():
            for s in d.decode(stream['stream'], self.get_stream_encodings(stream)):
                if all((x == s['stream'][0] for x in s['stream'])):
                    continue
                s['parent'] = stream
                yield s
                for sub in self.get_streams(s, depth+1):
                    yield sub

    def get_stream_encodings(self, stream):
        """
        Helper method to return the current encoding chain as a dotted string.
        
        :param stream: Stream object to return encoding for.
        :return: str of enodings applied or None.
        """
        steps = []
        s = stream
        while True:
            if s['encoding']:
                steps.append(s['encoding'])
            if not s['parent']:
                break
            s = s['parent']
        if steps:
            return '.'.join(steps[::-1])
        return None
    
    def scan_buffer(self, buf, log=False):
        """
        Entry point to scan a buffer of bytes for embedded objects.

        :param buf: Bytes object containing data to scan.
        :return: Generator of matched result dicts.
        """
        dedupe = set()
        # If input is a mime encoded format like email or mht, quoted-printable can screw with plain
        # extraction of url's, etc... we don't really want to implement as a sub stream or we'll get
        # the broken urls, etc. from the original content still.  It can also create havoc with b64 data.
        # Force conversion upfront.  TODO: how do we correct file/match offsets when this triggers?
        if b'quoted-printable' in buf[:10000]: # and all(c < 0x7f for c in buf):
            if log:
                print("Forcing quoted-printable decode of input")
            buf = quopri.decodestring(buf)
            # TODO: invetigate if this could break email with mix of quopri and base64?
        if log:
            print("Scanning sub-streams and their permutations...")
            it = progressbar.progressbar(self.get_streams(buf), redirect_stdout=True)
        else:
            it = self.get_streams(buf)

        for stream in it:
            #print("Stream Encoding %s %s %d %s" % (stream['stream_id'], self.get_stream_encodings(stream) or 'none', len(stream['stream']), stream['stream'][:50]))
            for p in self.payloads.values():
                for res in chain(self.check_stream_plaintext(p, stream), p.deob(stream['stream'])):

                    # only include if deemed 'interesting'
                    if self.embedded_only and not res['offset'] \
                        and not stream['encoding'] \
                        and res['scheme'] == 'plain':
                        continue
                    h = md5(res['content']).hexdigest()
                    if stream['encoding']:
                        res['encoding'] = self.get_stream_encodings(stream)
                        res['stream_offset'] = res['offset']
                        res['offset'] = stream['offset']
                        res['stream_id'] = stream['stream_id']
                    dedupe_key = (res['scheme'], res['key'], stream['encoding'],
                        res.get('stream_offset'), res['offset'], h)
                    if dedupe_key in dedupe:
                        continue
                    # ignore no-op encoding of hex/reverse/rol4
                    if res.get('encoding') and 'base16' in res['encoding'] \
                            and 'reverse' in res['encoding'] \
                            and res['scheme'] == 'rol' and res['key'] == b'\x04':
                        continue
                    dedupe.add(dedupe_key)
                    yield res

    def check_stream_plaintext(self, payload, stream):
        """
        Check if the stream is a payload, without any further deobfuscation.

        Normally, paylaods would get picked up by ^ 0x0 anyway but
        this may find additional embedded objects that don't match
        known patterns (eg. PE with dos stub stripped, urls with mixed casing).
        """
        if not stream['encoding']:
            return
        start, end = payload.validator.validate(stream['stream'])
        if end > 0:
            res = {
                'offset': start,
                'content': stream['stream'],
                'scheme': 'plain',
                'payload_type': payload.name,
                'pattern': b'',
                'length': end - start,
                'key': None,
                'modifiers': {},
            }
            if end < len(stream['stream']):
                res['content_with_overlay'] = stream['stream']
                res['length_with_overlay'] = len(stream['stream'])
            yield res

    def scan_path(self, path):
        """
        Scan the file at the supplied path.
        """
        with open(path, 'rb') as tmp:
            for x in self.scan_buffer(tmp.read()):
                yield x


@begin.start(auto_convert=True)
def main(extract=False, output_dir=os.getcwd(), *files):
    """
    Command-line entrypoint to scan a list of files.
    
    Will scan any filepaths specified and print details of
    detected embedded payloads.
    
    :param extract: Whether to write out deobfuscated payloads to file.
    :param output_dir: Where to write out deobfuscated payloads.
    :param files: List of files to scan.
    """
    scanner = Scanner()
    for x in files:
        try:
            buf = open(x, 'rb').read()
            start = timeit.default_timer()
            print("%s (%d bytes):" % (x, len(buf)))
            for d in scanner.scan_buffer(buf, log=True):
                print("\tOffset: 0x%08x Length: %i" % (d['offset'], d['length']))
                if d.get('encoding'):
                    print("\tStream Id: %s Offset: 0x%08x" % (d['stream_id'], d['stream_offset']))
                    print("\tStream Encoding: %s" % d['encoding'])
                key = d['key']
                if key:
                    key = '%s' % binascii.hexlify(d['key']).decode()
                print("\tPayload: %s" % d['payload_type'])
                print("\tScheme: %s %s %s" % (d['scheme'], key, d['modifiers']))
                if len(d['content']) > 200:
                    print("\tSample: %s\n" % repr(d['content'][:35]))
                else:
                    print("\tContent: %s\n" % repr(d['content']))
                if extract:
                    with open(os.path.join(output_dir,
                                get_filename(x, d)), 'wb') as tmp:
                        tmp.write(d['content'])
                    if d.get('content_with_overlay'):
                        with open(os.path.join(output_dir,
                                    get_filename(x, d, True)), 'wb') as tmp:
                            tmp.write(d['content_with_overlay'])

            # print("\tDuration: %02d ms\n" % ((timeit.default_timer() - start) * 1000))
        except Exception as ex:
            print("Unable to process %s: %s" % (x, str(ex)))
            raise ex




# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

import re

from malcarve.schemes import rol, xor


MAX_STRING_LEN = 512


def deob(buf, schemes):
    scanner = StringScanner('url', 12, 256, 8, [])
    scanner.add_scheme(xor.XORPatternFinder('xor', 'xor key patterns', [], [], []))
    scanner.add_scheme(rol.ROLFinder('rol', 'rotate', [], [], []))
    scanner.add_pattern((0, b'http://'))

    for x in scanner.deob(buf):
        yield x


class URLValidator(object):
    # at least 2 domains deep and optionally any valid chars after (including ports).. no checking for user:pass
    pattern = re.compile(rb'((ftp|http)s?://[a-z\-0-9]{1,256}\.[a-z\-0-9]{1,256}([a-z\-0-9\_./:\%\?\#\=\+\~])*)', re.I)
    # PE certificate urls will often run into ascii numbers
    fixers = (
        b".cer0",
        b".com0",
        b".com1",
        b".com/0",
        b".crl0",
        b".crt0",
        b".htm0",
        b".html0",
        b"/ca10",
        b"/CPS0",
        b"/cps0",
        b"/DPM0",
        b"/policy/0",
        b"/repository0",
        b"/rpa0",
        b"/ts0",
        # happens with add encoding list of urls (comma gets translated)
        b"+",
    )

    def __init__(self, checks):
        self.checks = checks

    def validate(self, buf):
        """
        Simple checks on whether this looks like a valid URL or not.
        """
        # strip nulls for wide char, etc.
        #buf = bytes([x for x in buf if x != 0])
        # TODO: split if multiple urls run into each other
        valid = URLValidator.pattern.match(buf)
        if valid:
            url = valid.group(1)
            if url.endswith(URLValidator.fixers):
                return 0, len(url)-1
            if url[:-1].endswith(URLValidator.fixers):
                return 0, len(url)-2
            return 0, len(url)
        return -1, -1


class UAValidator(object):
    pattern = re.compile(rb'(Mozilla/\d([ A-Za-z\-0-9\_./:\%\?\#\=\+\~\(\);,]){2,256})')

    def __init__(self, checks):
        self.checks = checks

    def validate(self, buf):
        """
        Simple checks on whether this looks like a valid URL or not.
        """
        # strip nulls for wide char, etc.
        #buf = bytes([x for x in buf if x != 0])
        valid = UAValidator.pattern.match(buf)
        if valid:
            ua = valid.group(1)
            return 0, len(ua)
        return -1, -1


class NoopValidator(object):
    def __init__(self, checks):
        pass

    def validate(self, buf):
        return 0, len(buf)


class StringScanner(object):
    """
    Scanning class to find embedded/obfuscated strings.
    """
    def __init__(self, name, scan_min, scan_max, min_size, validation, **kwargs):
        self.name = name
        self.scan_min = scan_min
        self.scan_max = scan_max
        self.min_size = min_size
        # a bit of a hack as can get a list
        if not validation:
            self.validator = NoopValidator(validation)
        elif 'url' in validation:
            self.validator = URLValidator(validation)
        elif 'useragent' in validation:
            self.validator = UAValidator(validation)
        else:
            raise Exception('Unknown validation for string')
        self.schemes = []
        self.patterns = []

    def add_scheme(self, scheme):
        """
        Add obfuscation scheme to search for.
        
        :param scheme: Scheme scanning object
        """
        self.schemes.append(scheme)

    def add_pattern(self, pattern):
        """
        Add plaintext pattern to search for.
        
        :param pattern: Tuple of offset, string pattern
        """
        self.patterns.append(pattern)

    def deob(self, buf):
        """
        Deobfuscate any embedded strings.
        
        :param buf: Byte string buffer to scan
        :return: Generator of deobfuscated details (dict)
        """
        for offset, pat in self.patterns:
            for scheme in self.schemes:
                nexti = 0
                while True:
                    match = scheme.locate(buf[nexti:len(buf) - self.min_size], pat, offset,
                                          None, None)
                    if not match:
                        break
                    match['offset'] += nexti
                    nexti = match['offset'] + 1
                    match = scheme.extract(buf, match, min(MAX_STRING_LEN, len(buf)-match['offset']))
                    start, end = self.validator.validate(match['content'])
                    size = end - start
                    if size > 0:
                        match['payload_type'] = self.name
                        match['pattern'] = pat
                        match['content'] = match['content'][start:end]
                        match['length'] = size
                        yield match




# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

from base64 import b64decode
from binascii import unhexlify
import re


class ChrDecoder(object):
    """
    A stream decoder that will attempt to find and decode
    runs of character code conversions like:
    - Chr(111),Chr(73)
    - 123,45,67
    """
    def __init__(self, name):
        self.name = name
        self.pattern = re.compile(rb'(Chr\()?\d{1,3}\)?(\s{,5}[\s\-\,\&\|O\%\^\.\;]\s{,5}(Chr\()?\d{1,3}\)?){9,}')
        self.extract = re.compile(rb'(\d{1,3})')
    
    def decode(self, buf, encoding=None):
        for i, x in enumerate(re.finditer(self.pattern, buf)):
            try:
                buf = bytes((int(c.group(1)) for c in re.finditer(self.extract, x.group(0))))
            except Exception as ex:
                continue
            yield {'encoding': 'charcodes',
                   'stream_id': i,
                   'offset': x.start(),
                   'stream': buf,
                   }


class HexDecoder(object):
    """
    A stream decoder that will attempt to extract and unhexlify
    any base16 streams present in scanned buffers.
    """
    def __init__(self, name):
        self.name = name
        # allow separator char like space, comma, %
        self.pattern = re.compile(rb'([a-f0-9]{1,}[\,\^\%]?\s*){10,}|([A-F0-9]{2,}[\,\^\%]?\s*){10,}')

    def decode(self, buf, encoding=None):
        i = 0
        joined = b''
        first_offset = 0
        last_offset = 0
        for x in re.finditer(self.pattern, buf):
            try:
                m = x.group(0).lower() \
                    .replace(b',', b'') \
                    .replace(b'%', b'') \
                    .replace(b'^', b'')
                m = re.sub(rb'\s+', b'', m)
                if not joined or x.start() - last_offset <= 50:
                    joined += m
                else:
                    joined = m
                    first_offset = x.start()
                if not i:
                    first_offset = x.start()
                last_offset = x.end()
                i += 1

                if len(m) % 2:
                    m = m[:-1]
                s = unhexlify(m)
                yield {'encoding': 'base16',
                       'stream_id': i,
                       'offset': x.start(),
                       'stream': s,
                       }
            except Exception as ex:
                pass
        # try joining the found blobs?
        if i <= 1:
            return
        try:
            joined = unhexlify(joined)
            yield {'encoding': 'base16',
                   'stream_id': 0,
                   'offset': first_offset,
                   'stream': joined,
                   }
        except Exception:
            pass


class B64Decoder(object):
    """
    A stream decoder that will attempt to identify and
    extract any base64 encoded data contained in scanned buffers.
    """

    UPPER = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    def __init__(self, name):
        self.name = name
        self.pattern = re.compile(rb'([a-zA-Z0-9+/=]{20,}\s*)+')

    def decode(self, buf, encoding=None):
        i = 0
        joined = b''
        first_offset = 0
        for x in re.finditer(self.pattern, buf):
            # trying to filter out straight hex encoded
            # all our patterns should result in mix of casing
            if not any(c in B64Decoder.UPPER for c in x.group(0)[:100]):
                continue
            try:
                chunk = re.sub(rb'\s+', b'', x.group(0))
                rem = len(chunk) % 4
                if rem == 3:
                    chunk += b'='
                elif rem == 2:
                    chunk += b'=='
                elif rem == 1:
                    # extra char (or multiple missing) somewhere
                    chunk = chunk[:-1]
                s = b64decode(chunk)
                if not joined.endswith(b'='):
                    joined += re.sub(rb'\s+', b'', x.group(0))
                    i += 1
                    first_offset = x.start()

                yield {'encoding': 'base64',
                       'stream_id': i,
                       'offset': x.start(),
                       'stream': s,
                       }
            except Exception as ex:
                pass

        # try joining the found blobs?
        if i <= 1:
            return
        try:
            joined = base64.decode(joined)
            #yield {'encoding': 'base64',
            #       'stream_id': 0,
            #       'offset': first_offset,
            #       'stream': joined,
            #       }
        except:
            pass

class VariableDecoder(object):
    """
    A stream decoder that will attempt to identify and
    extract string variable definition/concatenations in scripts.
    """
    def __init__(self, name):
        self.name = name
        self.varpattern = re.compile(rb'([a-zA-Z0-9\s+&=_\[\(\]\)]{2,30}(\"|\')([a-zA-Z0-9+/=]{16,})(\"|\')\s+\;?){2,}')
        self.stringpattern = re.compile(rb'(\'|\")([a-zA-Z0-9+/=]+)(\'|\")')

    def decode(self, buf, encoding=None):
        for i, x in enumerate(self.varpattern.finditer(buf, re.DOTALL)):
            b = b''.join(z[1] for z in self.stringpattern.findall(x.group(0)))
            # ughh.. did we just combine multi b64 strings?  they won't decode together if padded
            while b:
                eql = b.find(b'=')
                if eql >= 0:
                    if len(b) > eql+1 and b[eql+1] == 61:
                        eql += 1
                    yield {'encoding': 'vars',
                           'stream_id': i,
                           'offset': x.start(),
                           'stream': b[:eql+1],
                           }
                    b = b[eql+1:]
                    continue
                yield {'encoding': 'vars',
                       'stream_id': i,
                       'offset': x.start(),
                       'stream': b,
                       }
                break



# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

import re
import zlib

class DeflateDecoder(object):
    """
    Generic decoder for zlib/deflate as used in several file formats.
    Eg. PDF, compressed flash, image formats, gzip.
    For complete decoding will probably want to redo as file format
    specific decoders that can parse their formats accurately.
    
    https://tools.ietf.org/html/rfc1950
    """
    def __init__(self, name):
        self.name = name
        # 0x4889 - 4K Window size common in PDF Deflate Filter
        # 0x7801 - 32K Window Low Compression
        # 0x789C - 32K Window Default Compression
        # 0x78DA - 32K Window Best Compression
        # TODO: this is probably not exhaustive and there aren't really fixed bytes
        self.pattern = re.compile(
            rb'((\x48\x89)|(\x78\x01)|(\x78\x9c)|(\x78\xda)|(\xec\xfd)|(\xed\x7c)' +
            rb'|(\xe4\x5c)|(\x7c\x92)|(\x9c\x53)|(\x8c\x8f)|(\xec\x9d)|(\xec\x59)' +
            rb'|(\xc4\x54)|(\xc4\x53)|(\x84\xd0)|(\x9d\x54)|(\xcc\x58)|(\xac\x95)' +
            rb'|(\x8c\x92)|(\xc4\x96))'
        )
        self.min_size = 12

    def decode(self, buf, encoding=None):
        # don't enumerate() as we only want to count successful deflate streams
        # likely lots of false positives with 2 byte pattern.. should tighten up
        i = 0
        for x in re.finditer(self.pattern, buf):
            try:
                # try with zlib header, 0 = auto detect wbits
                yield {'encoding': 'deflate',
                       'stream_id': i,
                       'offset': x.start(),
                       'stream': zlib.decompress(buf[x.start():], 0),
                       }
                i += 1
            except Exception:
                # try without zlib header 'raw'
                try:
                    yield {'encoding': 'deflate',
                           'stream_id': i,
                           'offset': x.start(),
                           'stream': zlib.decompress(buf[x.start():], -15),
                           }
                    i += 1
                except Exception:
                    # give up
                    pass
    def validate(self, buf):
        if len(buf) < self.min_size:
            raise Exception("Too Small")
        return buf



# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import platform
import sys

from ctypes import byref, cdll, create_string_buffer
from ctypes import c_bool, c_size_t, c_int32, pointer

import malcarve

def load_library(library_name):
    """
    Load the ibrary with the given name using ctypes.
    Code lifted from https://github.com/sptonkin/fuzzyhashlib
    """
    # Figure out architecture.
    if platform.architecture()[0].startswith("64"):
        arch = "x86_64"
    else:
        arch = "x86_32"

    # Figure out OS info.
    if sys.platform.startswith("linux"):
        os_platform = "linux"
        extension = "so"
    elif sys.platform.startswith("darwin"):
        os_platform = "darwin"
        extension = "dylib"
    elif sys.platform.startswith("win"):
        os_platform = "windows"
        extension = "dll"
    else:
        raise Exception("Unsupported platform - %s" % sys.platform)

    library_filename = "%s.%s" % (library_name, extension)
    library_path = os.path.join(os.path.dirname(malcarve.__file__),
                                "ext",
                                "lib",
                                os_platform,
                                arch,
                                library_filename)
    return cdll.LoadLibrary(library_path)

x = load_library('xorpatterns')

def xor(buf, key, key_step=0, offset=0,
        null_preserve=False, rolling=False, decode=False):
    """
    XOR the buffer with the supplied key.
    
    :param buf: Byte string to obfscuated/deobfuscate.
    :param key: Key string to use for XOR.
    :param key_step: Integer to modify the key by after each XOR.
    :param offset: Offset from start of file to this buffer.
    :param null_preserve: Ignore Null and XOR values.
    :param rolling: Rolling XOR where previous output is next key
    :param decode: Are we encoding or decoding (only affects some modifiers)
    :return: XOR'ed buffer
    """
    outbuf = create_string_buffer(len(buf))
    if key_step:
        x.xor_countup(buf, c_size_t(len(buf)),
                      key, c_size_t(len(key)),
                      key_step, c_size_t(offset), c_bool(null_preserve),
                      outbuf)
    elif rolling:
        x.xor_rolling(buf, c_size_t(len(buf)),
                      key, c_size_t(len(key)),
                      decode, outbuf)
    else:
        x.xor(buf, c_size_t(len(buf)),
              key, c_size_t(len(key)),
              c_bool(null_preserve),
              outbuf)
    return outbuf.raw

def keypattern(pattern, original):
    """
    Given a buffer XOR'ed with a plaintext pattern, attempt
    to identify a repeating XOR key pattern.
    
    :param pattern: The XOR'ed buffer
    :param original: Original plaintext used
    :return: Dict with details of pattern or empty if none identified 
    """
    keybuf = create_string_buffer(64)
    keysize = c_size_t()
    np = c_bool()
    step = c_int32()
    found = x.keypattern(pattern,
                         original, 0, c_size_t(min(8, int(len(pattern)/4))),
                         keybuf, pointer(keysize),
                         pointer(np), pointer(step))
    if found:
        res = {
            'scheme': 'xor',
            'key': keybuf.raw[:keysize.value],
            'keysize': keysize.value,
        }
        modifiers = {}
        if np:
            modifiers['null_preserve'] = True
        if step:
            modifiers['step'] = step.value
        res['modifiers'] = modifiers
        return res
    return {}

def findxor(buf, pattern, patoffset, nulloffset=None, header=None):
    """
    Attempt to identify a XOR'ed pattern in the supplied buffer.
    
    :param buf: Byte string to scan
    :param pattern: Byte string pattern to search for
    :param patoffset: Integer offset where pattern expected relative to start
    :param nulloffset: Alternative location to test for null bytes
    :param header: Header pattern hint if not searching at offset 0
    :return: Dict with details of located scheme or empty if none found
    """
    keybuf = create_string_buffer(64)
    keysize = c_size_t()
    np = c_bool()
    step = c_int32()
    roll = c_bool()
    offset = c_size_t()

    found = x.findxor(buf, c_size_t(len(buf)),
                      pattern, c_size_t(len(pattern)), c_size_t(patoffset),
                      pointer(offset), keybuf, pointer(keysize),
                      pointer(np), pointer(step), pointer(roll))
    if found:
        res = {
            'offset': offset.value,
            'scheme': 'xor',
            'key': keybuf.raw[:keysize.value],
            'keysize': keysize.value,
        }
        modifiers = {}
        if np:
            modifiers['null_preserve'] = True
        if step:
            modifiers['step'] = step.value
        if roll:
            modifiers['rolling'] = True

        res['modifiers'] = modifiers
        # TODO: move into c code?
        # we can miss null preservation if not long enough runs in pattern
        if nulloffset and not np:
            check_location = offset.value + nulloffset
            if buf[check_location:check_location + keysize.value] == b'\x00' * keysize.value:
                modifiers['null_preserve'] = True
        # rolling start key value likely incorrect if not at offset 0 :(
        # try to fix with expected/known header bytes
        if roll and patoffset and header and len(header) >= keysize.value:
            res['key'] = bytes([c ^ buf[offset.value + i]
                                  for i, c in enumerate(header) if i < keysize.value])
        return res
    return {}


class XORPatternFinder(object):
    """
    Finds XOR patterns in byte streams and deobfuscates their content.
    """
    def __init__(self, name, description, schemes, modifiers, max_keysize, **kwargs):
        self.name = name
        self.description = description
        # TODO.. maybe remove?
        self.enabled_schemes = schemes
        self.enabled_modifiers = modifiers
        self.max_keysize = max_keysize

    def locate(self, buf, pattern, pattern_offset,
               alternate_nulls_offset=None, probable_header=None):
        """
        Attempt to find an obfuscated instance of the pattern in the
        supplied buffer.
        
        :param buf: Byte string buffer to scan
        :param pattern: Plaintext pattern to search for
        :param pattern_offset: Offset of pattern from start of payload/file
        :param alternate_nulls_offset: Alternate location to scan for null pattern
        :param probable_header: Workaround for rolling xors where pattern not at 0
        :return: Dictionary of located pattern or empty if not found
        """
        return findxor(buf, pattern, pattern_offset, alternate_nulls_offset, probable_header)

    def extract(self, buf, match, size=None):
        """
        Extract the deobfuscated content of the matched scheme from the buffer.
        
        :param buf: Byt string buffer that was matched
        :param match: Dictionary of match details to extract
        :return: Dictionary of match with deobfsucated content
        """
        if not size:
            end = len(buf)
        else:
            end = match['offset'] + size
        content = xor(buf[match['offset']:end],
                          match['key'],
                          null_preserve=match['modifiers'].get('null_preserve'),
                          key_step=match['modifiers'].get('step', 0),
                          rolling=match['modifiers'].get('rolling'),
                          decode=True)
        match['content'] = content
        match['length'] = len(content)

        # plaintext special case
        if match['scheme'] == 'xor' and match['key'] == b'\x00' \
            and (not match['modifiers'] \
                 or match['modifiers'] == {'null_preserve': True}):
            match['scheme'] = 'plain'
            match['key'] = None
            match['keysize'] = 0
            match['modifiers'] = {}
        return match



# Malcarve - Obfuscated payload extractor for malware samples
# Copyright (C) 2016 Steve Henderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import unicode_literals

import re

def add(buf, key, encode=True):
    """
    Single byte add/sub.
    Encode to add, decode to sub.
    """
    if type(key) != int:
        key = ord(key)
    if not encode:
        key *= -1
    if type(buf) != bytes:
        buf = buf.encode()
    return bytes(((x + key) % 256 for x in buf))

def findadd(buf, pattern, pattern_offset):
    """
    Simple brute forcer for single byte add obfuscations.
    """
    # should be quicker to modify pattern not buf
    addpats = []
    for i in range(1, 256):
        # assume we'll also be doing xor and skip 0x80 as equivalent
        if i == 0x80:
            continue
        addpats.append(re.escape(add(pattern, i)))
    pat = b'(%s)' % b'|'.join(addpats)
    for x in re.finditer(pat, buf[pattern_offset:]):
        i = addpats.index(re.escape(x.group(0))) + 1
        if i >= 0x80:
            i += 1
        return {'scheme': 'add',
               'key': bytes([i]),
               'keysize': 1,
               'offset': x.start(),
               'modifiers': {},
               }

class AddFinder(object):
    """
    Finds simple, single byte additions in byte streams and deobfuscates their content.
    """
    def __init__(self, name, description, schemes, modifiers, max_keysize, **kwargs):
        self.name = name
        self.description = description
        # TODO.. maybe remove?
        self.enabled_schemes = schemes
        self.enabled_modifiers = modifiers
        self.max_keysize = max_keysize

    def locate(self, buf, pattern, pattern_offset,
               alternate_nulls_offset=None, probable_header=None):
        """
        Attempt to find an obfuscated instance of the pattern in the
        supplied buffer.
        
        :param buf: Byte string buffer to scan
        :param pattern: Plaintext pattern to search for
        :param pattern_offset: Offset of pattern from start of payload/file
        :param alternate_nulls_offset: Alternate location to scan for null pattern
        :param probable_header: Workaround for rolling xors where pattern not at 0
        :return: Dictionary of located pattern or empty if not found
        """
        return findadd(buf, pattern, pattern_offset)

    def extract(self, buf, match, size=None):
        """
        Extract the deobfuscated content of the matched scheme from the buffer.
        
        :param buf: Byt string buffer that was matched
        :param match: Dictionary of match details to extract
        :return: Dictionary of match with deobfsucated content
        """
        if not size:
            end = len(buf)
        else:
            end = match['offset'] + size

        content = add(buf[match['offset']:],
                          match['key'],
                          encode=False)
        match['content'] = content
        match['length'] = len(content)

        return match
